# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
import numpy as np

__author__ = "Sudarsan Surendralal"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class Dos(object):
    """
    The DOS class stores all information to store and retrieve the total and resolved density of states from an
    electronic structure calculation.

    Args:
        n_bins (int): Number of histogram bins required to calculate the DOS
        es_obj: The pyiron_atomistics.objects.waves.core.ElectronicStructure instance for which the DOS has to be computed
        eigenvalues (list/numpy.ndarray): If es-obj is None, the eigenvalues could be specified as a list

    """

    def __init__(self, n_bins=100, es_obj=None, eigenvalues=None, bin_density=None):
        self.orbital_dict = {"s": [0], "p": [1, 2, 3], "d": [4, 5, 6, 7, 8]}
        self.n_bins = n_bins
        self.es_obj = es_obj
        self.t_dos = list()
        self.energies = list()

        if es_obj is not None:
            eig_vals = self.es_obj.eigenvalues
        else:
            eig_vals = eigenvalues
        for eig_val in eig_vals:
            dos_min = np.min(eig_val)
            dos_max = np.max(eig_val)
            if bin_density is not None:
                n_bins = int((dos_max - dos_min) * bin_density)
            else:
                n_bins = self.n_bins
            t_dos, energies = np.histogram(eig_val, bins=int(n_bins), density=True)
            self.t_dos.append(t_dos)
            self.energies.append(energies)
        self.energies = [
            energies[1:] - ((energies[1] - energies[0]) / 2.0)
            for energies in self.energies
        ]

    def plot_total_dos(self, **kwargs):
        """
        Plots the total DOS

        Args:
            **kwargs: Variables for matplotlib.pylab.plot customization (linewidth, linestyle, etc.)

        Returns:
            matplotlib.pylab.plot
        """
        try:
            import matplotlib.pylab as plt
        except ImportError:
            import matplotlib.pyplot as plt
        fig = plt.figure(1, figsize=(6, 4))
        ax1 = fig.add_subplot(111)
        ax1.set_xlabel("E (eV)", fontsize=14)
        ax1.set_ylabel("DOS", fontsize=14)
        for i, energies in enumerate(self.energies):
            plt.fill_between(
                energies, self.t_dos[i], label="spin {}".format(i), **kwargs
            )
        plt.legend()
        return plt

    def plot_orbital_resolved_dos(self, **kwargs):
        """
        Plots the orbital resolved DOS

        Args:
            **kwargs: Variable for matplotlib.pylab.plot customization (linewidth, linestyle, etc.)

        Returns:
            matplotlib.pylab.plot
        """
        try:
            import matplotlib.pylab as plt
        except ImportError:
            import matplotlib.pyplot as plt
        if not (self.es_obj.grand_dos_matrix is not None):
            raise NoResolvedDosError(
                "Can not plot the orbital resolved dos since resolved dos values are not"
                " available"
            )
        plot = self.plot_total_dos()
        for spin in range(len(self.energies)):
            for key, val in self.orbital_dict.items():
                r_dos = self.get_orbital_resolved_dos(val)
                plt.plot(
                    self.energies, r_dos, label=key + "spin {}".format(spin), **kwargs
                )
        plot.legend()
        return plot

    def get_spin_resolved_dos(self, spin_indices):
        """
        Gives the dos contribution of a given indices of spin as arranged in the
        pyiron_atomistics.objects.waves.ElectronicStructure instance.

        Args:
            spin_indices (list/numpy.ndarray): The index/indices of the spins for which the dos contribution is required

        Returns:
            numpy.ndarray: The required dos

        """
        if not (self.es_obj.grand_dos_matrix is not None):
            raise NoResolvedDosError(
                "Can not get the spin resolved dos since resolved dos values are not"
                " available"
            )

        grand_sum = np.sum(self.es_obj.grand_dos_matrix)
        tot_val = self.es_obj.grand_dos_matrix.copy() / grand_sum
        n_spin, n_kpts, n_bands, _, _ = np.shape(tot_val)
        k = 0
        b = 0
        r_dos = np.zeros_like(self.t_dos[spin_indices])
        w_dos = np.zeros_like(self.t_dos[spin_indices])
        eigenvalues = self.es_obj.eigenvalues[spin_indices]
        for e in eigenvalues:
            weight = np.sum(tot_val[spin_indices, k, b, :, :])
            weight_sum = np.sum(tot_val[:, k, b, :, :])
            if b < n_bands - 1:
                b += 1
            else:
                b = 0
                k += 1
            index = (
                len(self.energies[spin_indices][self.energies[spin_indices] < e]) - 1
            )
            if index >= 0:
                r_dos[index] = r_dos[index] + weight
                w_dos[index] = w_dos[index] + weight_sum
            else:
                r_dos[0] = r_dos[0] + weight
                w_dos[0] = w_dos[0] + weight_sum
        ind_0 = np.argwhere(w_dos < 1e-8).flatten()
        ind_1 = np.argwhere(w_dos >= 1e-8).flatten()
        r_dos[ind_1] /= w_dos[ind_1]
        r_dos[ind_0] = 0.0
        return r_dos * self.t_dos[spin_indices]

    def get_spatially_resolved_dos(self, atom_indices, spin_indices=0):
        """
        Gives the dos contribution of a given indices of atoms as arranged in the
        pyiron_atomistics.objects.waves.ElectronicStructure instance.

        Args:
            atom_indices (list/numpy.ndarray): The index/indices of the atoms for which the dos contribution is required
            spin_indices (list/numpy.ndarray): The index/indices of the spins for which the dos contribution is required

        Returns:
            numpy.ndarray: The required dos

        """
        if not (self.es_obj.grand_dos_matrix is not None):
            raise NoResolvedDosError(
                "Can not get the spatially resolved dos since resolved dos values are not"
                " available"
            )
        grand_sum = np.sum(self.es_obj.grand_dos_matrix)
        tot_val = self.es_obj.grand_dos_matrix.copy() / grand_sum
        _, n_kpts, n_bands, _, _ = np.shape(tot_val)
        k = 0
        b = 0
        r_dos = np.zeros_like(self.t_dos[spin_indices])
        w_dos = np.zeros_like(self.t_dos[spin_indices])
        eigenvalues = self.es_obj.eigenvalues[spin_indices]
        for e in eigenvalues:
            weight = np.sum(tot_val[spin_indices, k, b, atom_indices, :])
            weight_sum = np.sum(tot_val[spin_indices, k, b, :, :])
            if b < n_bands - 1:
                b += 1
            else:
                b = 0
                k += 1
            index = (
                len(self.energies[spin_indices][self.energies[spin_indices] < e]) - 1
            )
            if index >= 0:
                r_dos[index] = r_dos[index] + weight
                w_dos[index] = w_dos[index] + weight_sum
            else:
                r_dos[0] = r_dos[0] + weight
                w_dos[0] = w_dos[0] + weight_sum
        ind_0 = np.argwhere(w_dos < 1e-8).flatten()
        ind_1 = np.argwhere(w_dos >= 1e-8).flatten()
        r_dos[ind_1] /= w_dos[ind_1]
        r_dos[ind_0] = 0.0
        return r_dos * self.t_dos[spin_indices]

    def get_orbital_resolved_dos(self, orbital_indices, spin_indices=0):
        """
        Gives the dos contribution of a given indices of orbitals as arranged in the
        pyiron_atomistics.objects.waves.ElectronicStructure instance.

        Args:
            orbital_indices (list/numpy.ndarray): The index/indices of the orbitals for which the dos contribution is required
            spin_indices (list/numpy.ndarray): The index/indices of the spins for which the dos contribution is required

        Returns:
            numpy.ndaray: The required dos

        """
        if not (self.es_obj.grand_dos_matrix is not None):
            raise NoResolvedDosError(
                "Can not get the orbital resolved dos since resolved dos values are not"
                " available"
            )
        grand_sum = np.sum(self.es_obj.grand_dos_matrix)
        tot_val = self.es_obj.grand_dos_matrix.copy() / grand_sum
        _, n_kpts, n_bands, _, _ = np.shape(tot_val)
        k = 0
        b = 0
        r_dos = np.zeros_like(self.t_dos)[spin_indices]
        w_dos = np.zeros_like(self.t_dos)[spin_indices]
        eigenvalues = self.es_obj.eigenvalues[spin_indices]
        for e in eigenvalues:
            weight = np.sum(tot_val[spin_indices, k, b, :, orbital_indices])
            weight_sum = np.sum(tot_val[spin_indices, k, b, :, :])
            if b < n_bands - 1:
                b += 1
            else:
                b = 0
                k += 1
            index = (
                len(self.energies[spin_indices][self.energies[spin_indices] < e]) - 1
            )
            if index >= 0:
                r_dos[index] = r_dos[index] + weight
                w_dos[index] = w_dos[index] + weight_sum
            else:
                r_dos[0] = r_dos[0] + weight
                w_dos[0] = w_dos[0] + weight_sum
        ind_0 = np.argwhere(w_dos < 1e-8).flatten()
        ind_1 = np.argwhere(w_dos >= 1e-8).flatten()
        r_dos[ind_1] /= w_dos[ind_1]
        r_dos[ind_0] = 0.0
        return r_dos * self.t_dos[spin_indices]

    def get_spatial_orbital_resolved_dos(
        self, atom_indices, orbital_indices, spin_indices=0
    ):
        """
        Gives the dos contribution of a given indices of atoms as well as orbitals as arranged in the
        pyiron_atomistics.objects.waves.ElectronicStructure instance.

        Args:
            atom_indices (list/numpy.ndarray): The index/indices of the atoms for which the dos contribution is required
            orbital_indices (list/numpy.ndarray): The index/indices of the orbitals for which the dos contribution is required
            spin_indices (list/numpy.ndarray): The index/indices of the spins for which the dos contribution is required

        Returns:
            numpy.ndaray: The required dos
        """
        if not (self.es_obj.grand_dos_matrix is not None):
            raise NoResolvedDosError(
                "Can not get the resolved dos since resolved dos values are not"
                " available"
            )
        grand_sum = np.sum(self.es_obj.grand_dos_matrix)
        tot_val = self.es_obj.grand_dos_matrix.copy() / grand_sum
        _, n_kpts, n_bands, _, _ = np.shape(tot_val)
        k = 0
        b = 0
        r_dos = np.zeros_like(self.t_dos)[spin_indices]
        w_dos = np.zeros_like(self.t_dos)[spin_indices]

        eigenvalues = self.es_obj.eigenvalues[spin_indices]
        for e in eigenvalues:
            weight = np.sum(
                [
                    np.sum(tot_val[spin_indices, k, b, atom_indices, o])
                    for o in orbital_indices
                ]
            )
            weight_sum = np.sum(tot_val[spin_indices, k, b, :, :])
            if b < n_bands - 1:
                b += 1
            else:
                b = 0
                k += 1
            index = (
                len(self.energies[spin_indices][self.energies[spin_indices] < e]) - 1
            )
            if index >= 0:
                r_dos[index] = r_dos[index] + weight
                w_dos[index] = w_dos[index] + weight_sum
            else:
                r_dos[0] = r_dos[0] + weight
                w_dos[0] = w_dos[0] + weight_sum
        ind_0 = np.argwhere(w_dos < 1e-8).flatten()
        ind_1 = np.argwhere(w_dos >= 1e-8).flatten()
        r_dos[ind_1] /= w_dos[ind_1]
        r_dos[ind_0] = 0.0
        return r_dos * self.t_dos


class NoResolvedDosError(Exception):
    """
    Raised when information on the resolved dos in unavailable
    """

    pass



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
import math
import warnings

import numpy as np
from pyiron_vasp.vasp.structure import (
    atoms_from_string,
    get_species_list_from_potcar,
)
from pyiron_vasp.dft.volumetric import VolumetricData

__author__ = "Sudarsan Surendralal"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class VaspVolumetricData(VolumetricData):
    """
    General class for parsing and manipulating volumetric static within VASP. The basic idea of the Base class is
    adapted from the pymatgen vasp VolumtricData class

    http://pymatgen.org/_modules/pymatgen/io/vasp/outputs.html#VolumetricData

    """

    def __init__(self):
        super(VaspVolumetricData, self).__init__()
        self.atoms = None
        self._diff_data = None
        self._total_data = None

    def from_file(self, filename, normalize=True):
        """
        Parsing the contents of from a file

        Args:
            filename (str): Path of file to parse
            normalize (boolean): Flag to normalize by the volume of the cell
        """
        try:
            self.atoms, vol_data_list = self._read_vol_data(
                filename=filename, normalize=normalize
            )
        except (ValueError, IndexError, TypeError):
            try:
                self.atoms, vol_data_list = self._read_vol_data_old(
                    filename=filename, normalize=normalize
                )
            except (ValueError, IndexError, TypeError):
                raise ValueError("Unable to parse file: {}".format(filename))
        if self.atoms is not None:
            self._total_data = vol_data_list[0]
            if len(vol_data_list) > 1:
                self._diff_data = vol_data_list[1]

    @staticmethod
    def _read_vol_data_old(filename, normalize=True):
        """
        Convenience method to parse a generic volumetric static file in the vasp like format.
        Used by subclasses for parsing the file. This routine is adapted from the pymatgen vasp VolumetricData
        class with very minor modifications. The new parser is faster

        http://pymatgen.org/_modules/pymatgen/io/vasp/outputs.html#VolumetricData.

        Args:
            filename (str): Path of file to parse
            normalize (boolean): Flag to normalize by the volume of the cell

        """
        if os.stat(filename).st_size == 0:
            warnings.warn("File:" + filename + "seems to be corrupted/empty")
            return None, None
        poscar_read = False
        poscar_string = list()
        dataset = list()
        all_dataset = list()
        dim = None
        dimline = None
        read_dataset = False
        ngrid_pts = 0
        data_count = 0
        atoms = None
        volume = None
        with open(filename, "r") as f:
            for line in f:
                line = line.strip()
                if read_dataset:
                    toks = line.split()
                    for tok in toks:
                        if data_count < ngrid_pts:
                            # This complicated procedure is necessary because
                            # vasp outputs x as the fastest index, followed by y
                            # then z.
                            x = data_count % dim[0]
                            y = int(math.floor(data_count / dim[0])) % dim[1]
                            z = int(math.floor(data_count / dim[0] / dim[1]))
                            dataset[x, y, z] = float(tok)
                            data_count += 1
                    if data_count >= ngrid_pts:
                        read_dataset = False
                        data_count = 0
                        all_dataset.append(dataset)
                elif not poscar_read:
                    if line != "" or len(poscar_string) == 0:
                        poscar_string.append(line)
                    elif line == "":
                        try:
                            atoms = atoms_from_string(poscar_string)
                        except ValueError:
                            pot_str = filename.split("/")
                            pot_str[-1] = "POTCAR"
                            potcar_file = "/".join(pot_str)
                            species = get_species_list_from_potcar(potcar_file)
                            atoms = atoms_from_string(
                                poscar_string, species_list=species
                            )
                        volume = atoms.get_volume()
                        poscar_read = True
                elif not dim:
                    dim = [int(i) for i in line.split()]
                    ngrid_pts = dim[0] * dim[1] * dim[2]
                    dimline = line
                    read_dataset = True
                    dataset = np.zeros(dim)
                elif line == dimline:
                    read_dataset = True
                    dataset = np.zeros(dim)
            if not normalize:
                volume = 1.0
            if len(all_dataset) == 0:
                warnings.warn("File:" + filename + "seems to be corrupted/empty")
                return None, None
            if len(all_dataset) == 2:
                data = {
                    "total": all_dataset[0] / volume,
                    "diff": all_dataset[1] / volume,
                }
                return atoms, [data["total"], data["diff"]]
            else:
                data = {"total": all_dataset[0] / volume}
                return atoms, [data["total"]]

    def _read_vol_data(self, filename, normalize=True):
        """
        Parses the VASP volumetric type files (CHGCAR, LOCPOT, PARCHG etc). Rather than looping over individual values,
        this function utilizes numpy indexing resulting in a parsing efficiency of at least 10%.

        Args:
            filename (str): File to be parsed
            normalize (bool): Normalize the data with respect to the volume (Recommended for CHGCAR files)

        Returns:
            pyiron.atomistics.structure.atoms.Atoms: The structure of the volumetric snapshot
            list: A list of the volumetric data (length >1 for CHGCAR files with spin)

        """
        if not os.path.getsize(filename) > 0:
            warnings.warn("File:" + filename + "seems to be empty! ")
            return None, None
        with open(filename, "r") as f:
            struct_lines = list()
            get_grid = False
            n_x = 0
            n_y = 0
            n_z = 0
            n_grid = 0
            n_grid_str = None
            total_data_list = list()
            atoms = None
            for line in f:
                strip_line = line.strip()
                if not get_grid:
                    if strip_line == "":
                        get_grid = True
                    struct_lines.append(strip_line)
                elif n_grid_str is None:
                    n_x, n_y, n_z = [int(val) for val in strip_line.split()]
                    n_grid = n_x * n_y * n_z
                    n_grid_str = " ".join([str(val) for val in [n_x, n_y, n_z]])
                    load_txt = np.genfromtxt(f, max_rows=int(n_grid / 5))
                    load_txt = np.hstack(load_txt)
                    if n_grid % 5 != 0:
                        add_line = np.genfromtxt(f, max_rows=1)
                        load_txt = np.append(load_txt, np.hstack(add_line))
                    total_data = self._fastest_index_reshape(load_txt, [n_x, n_y, n_z])
                    try:
                        atoms = atoms_from_string(struct_lines)
                    except ValueError:
                        pot_str = filename.split("/")
                        pot_str[-1] = "POTCAR"
                        potcar_file = "/".join(pot_str)
                        species = get_species_list_from_potcar(potcar_file)
                        atoms = atoms_from_string(struct_lines, species_list=species)
                    if normalize:
                        total_data /= atoms.get_volume()
                    total_data_list.append(total_data)
                elif atoms is not None:
                    grid_str = n_grid_str.replace(" ", "")
                    if grid_str == strip_line.replace(" ", ""):
                        load_txt = np.genfromtxt(f, max_rows=int(n_grid / 5))
                        load_txt = np.hstack(load_txt)
                        if n_grid % 5 != 0:
                            add_line = np.genfromtxt(f, max_rows=1)
                            load_txt = np.hstack(
                                np.append(load_txt, np.hstack(add_line))
                            )
                        total_data = self._fastest_index_reshape(
                            load_txt, [n_x, n_y, n_z]
                        )
                        if normalize:
                            total_data /= atoms.get_volume()
                        total_data_list.append(total_data)
            if len(total_data_list) == 0:
                warnings.warn(
                    "File:"
                    + filename
                    + "seems to be corrupted/empty even after parsing!"
                )
                return None, None
            return atoms, total_data_list

    @staticmethod
    def _fastest_index_reshape(raw_data, grid):
        """
        Helper function to parse volumetric data with x-axis as the fastest index into a 3D numpy array

        Args:
            raw_data (numpy.ndarray): Raw unprocessed volumetric data which is flattened
            grid (list/turple/numpy.ndarray): Sequence of the integer grid points [Nx, Ny, Nz]

        Returns:
            numpy.ndarray: A Nx $\times$ Ny $\times$ Nz numpy array

        """
        n_x, n_y, n_z = grid
        total_data = np.zeros((n_x, n_y, n_z))
        all_data = raw_data[0 : np.prod(grid)]
        all_indices = np.arange(len(all_data), dtype=int)
        x_indices = all_indices % n_x
        y_indices = all_indices / n_x % n_y
        y_indices = np.array(y_indices, dtype=int)
        z_indices = all_indices / (n_x * n_y)
        z_indices = np.array(z_indices, dtype=int)
        total_data[x_indices, y_indices, z_indices] = all_data
        return total_data

    @property
    def total_data(self):
        """
        numpy.ndarray: Total volumtric data (3D)
        """
        return self._total_data

    @total_data.setter
    def total_data(self, val):
        self._total_data = val

    @property
    def diff_data(self):
        """
        numpy.ndarray: Volumtric difference data (3D)
        """
        return self._diff_data

    @diff_data.setter
    def diff_data(self, val):
        self._diff_data = val

    def to_dict(self):
        volumetric_data_dict = {
            "TYPE": str(type(self)),
            "total": self.total_data,
        }
        if self.diff_data is not None:
            volumetric_data_dict["diff"] = self.diff_data
        return volumetric_data_dict



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
import re

from ase.atoms import Atoms
from ase.constraints import FixCartesian
import numpy as np
from collections import OrderedDict
from pyiron_vasp.dft.waves.electronic import ElectronicStructure
import defusedxml.cElementTree as ETree
from defusedxml.ElementTree import ParseError
import warnings


__author__ = "Sudarsan Surendralal"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class Vasprun(object):
    """
    This module is used to parse vasprun.xml files and store the data consistent with the pyiron input/output storage
    formats.

    Attributes:

        vasprun_dict (dict): Dictionary containing all information from the calculation parsed from the vasprun.xml
                            file. If you consider a simulation with N atoms and M ionic steps

                'positions' (numpy.ndarray): MxNx3 array containing all the relative positions
                'cell' (numpy.ndarray): Mx3x3 array containing all the size and shape of cells at every iteration point
                'forces' (numpy.ndarray): MxNx3 array containing all the forces in eV/A
                'total_energies' (numpy.ndarray): 1xM array containing all the total energies in eV
    """

    def __init__(self):
        self.vasprun_dict = dict()

    def from_file(self, filename="vasprun.xml"):
        """
        Parsing vasprun.xml from the working directory

        Args:
            filename (str): Path to the vasprun file
        """
        if not (os.path.isfile(filename)):
            raise AssertionError()
        try:
            self.parse_root_to_dict(filename)
        except ParseError:
            raise VasprunError(
                "The vasprun.xml file is either corrupted or the simulation has failed"
            )

    def parse_root_to_dict(self, filename):
        """
        Parses from the main xml root.
        """
        d = self.vasprun_dict
        d["scf_energies"] = list()
        d["scf_fr_energies"] = list()
        d["scf_0_energies"] = list()
        d["scf_dipole_moments"] = list()
        d["positions"] = list()
        d["cells"] = list()
        d["forces"] = list()
        d["total_energies"] = list()
        d["total_fr_energies"] = list()
        d["total_0_energies"] = list()
        d["kinetic_energies"] = list()
        d["stress_tensors"] = list()
        for _, leaf in ETree.iterparse(filename):
            if leaf.tag in ["generator", "incar"]:
                d[leaf.tag] = dict()
                for items in leaf:
                    d[leaf.tag] = self.parse_item_to_dict(items, d[leaf.tag])
            if leaf.tag in ["kpoints"]:
                d[leaf.tag] = dict()
                self.parse_kpoints_to_dict(leaf, d[leaf.tag])
            if leaf.tag in ["atominfo"]:
                d[leaf.tag] = dict()
                self.parse_atom_information_to_dict(leaf, d[leaf.tag])
            if leaf.tag in ["structure"] and "name" in leaf.keys():
                if "initialpos" in leaf.attrib["name"]:
                    d["init_structure"] = dict()
                    self.parse_structure_to_dict(leaf, d["init_structure"])
                elif "finalpos" in leaf.attrib["name"]:
                    d["final_structure"] = dict()
                    self.parse_structure_to_dict(leaf, d["final_structure"])
            if leaf.tag in ["calculation"]:
                self.parse_calc_to_dict(leaf, d)
            if leaf.tag in ["parameters"]:
                pass
                self.parse_parameters(leaf, d)
        d["cells"] = np.array(d["cells"])
        d["positions"] = np.array(d["positions"])
        # Check if the parsed coordinates are in absolute/relative coordinates. If absolute, convert to relative
        total_positions = d["positions"].flatten()
        if len(np.argwhere(total_positions > 1)) / len(total_positions) > 0.2:
            pos_new = d["positions"].copy()
            for i, pos in enumerate(pos_new):
                d["positions"][i] = np.dot(pos, np.linalg.inv(d["cells"][i]))
        d["forces"] = np.array(d["forces"])
        d["total_energies"] = np.array(d["total_energies"])
        d["total_fr_energies"] = np.array(d["total_fr_energies"])
        d["total_0_energies"] = np.array(d["total_0_energies"])
        if len(d["kinetic_energies"]) > 0:
            d["kinetic_energies"] = np.array(d["kinetic_energies"])
        else:
            del d["kinetic_energies"]
        d["scf_energies"] = d["scf_energies"]
        d["scf_dipole_moments"] = d["scf_dipole_moments"]
        d["scf_fr_energies"] = d["scf_fr_energies"]
        d["scf_0_energies"] = d["scf_0_energies"]
        d["stress_tensors"] = d["stress_tensors"]

    def parse_kpoints_to_dict(self, node, d):
        """
        Parses k-points data from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "kpoints"):
            raise AssertionError()
        for leaf in node:
            if leaf.tag == "generation":
                d[leaf.tag] = dict()
                d[leaf.tag]["scheme"] = leaf.attrib["param"]
                if d[leaf.tag]["scheme"] == "listgenerated":
                    line_mode_kpoints = list()
                    for item in leaf:
                        if item.tag == "v":
                            line_mode_kpoints.append(
                                self._parse_vector(item, vec_type=float)
                            )
                    d["line_mode_kpoints"] = np.array(line_mode_kpoints)
                else:
                    gen_vec = np.zeros((3, 3))
                    for item in leaf:
                        if item.tag == "v":
                            if item.attrib["name"] in ["divisions"]:
                                d[leaf.tag]["divisions"] = self._parse_vector(
                                    item, vec_type=int
                                )
                            if item.attrib["name"] in [
                                "genvec{}".format(i) for i in range(1, 4)
                            ]:
                                if item.attrib["name"] == "genvec1":
                                    gen_vec[0, :] = self._parse_vector(
                                        item, vec_type=float
                                    )
                                if item.attrib["name"] == "genvec2":
                                    gen_vec[1, :] = self._parse_vector(
                                        item, vec_type=float
                                    )
                                if item.attrib["name"] == "genvec3":
                                    gen_vec[2, :] = self._parse_vector(
                                        item, vec_type=float
                                    )
                            if item.attrib["name"] in ["shift", "usershift"]:
                                d[leaf.tag][item.attrib["name"]] = self._parse_vector(
                                    item, vec_type=float
                                )
                    d[leaf.tag]["genvec"] = np.array(gen_vec)
            if leaf.tag == "varray":
                if leaf.attrib["name"] == "kpointlist":
                    d["kpoint_list"] = self._parse_2d_matrix(leaf, vec_type=float)
                if leaf.attrib["name"] == "weights":
                    d["kpoint_weights"] = self._parse_2d_matrix(
                        leaf, vec_type=float
                    ).flatten()

    def parse_atom_information_to_dict(self, node, d):
        """
        Parses atom information from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "atominfo"):
            raise AssertionError()
        species_dict = OrderedDict()
        for leaf in node:
            if leaf.tag == "atoms":
                d["n_atoms"] = self._parse_vector(leaf)[0]
            if leaf.tag == "types":
                d["n_species"] = self._parse_vector(leaf)[0]
            if leaf.tag == "array":
                if leaf.attrib["name"] == "atomtypes":
                    for item in leaf:
                        if item.tag == "set":
                            for sp in item:
                                elements = sp
                                if elements[1].text in species_dict.keys():
                                    count = 1
                                    not_unique = True
                                    species_key = None
                                    while not_unique:
                                        species_key = "_".join(
                                            [elements[1].text, str(count)]
                                        )
                                        if species_key not in species_dict.keys():
                                            not_unique = False
                                        else:
                                            count += 1
                                    if species_key is not None:
                                        species_dict[species_key] = dict()
                                        species_dict[species_key]["n_atoms"] = int(
                                            elements[0].text
                                        )
                                        species_dict[species_key]["valence"] = float(
                                            elements[3].text
                                        )
                                else:
                                    species_key = elements[1].text
                                    species_dict[species_key] = dict()
                                    species_dict[species_key]["n_atoms"] = int(
                                        elements[0].text
                                    )
                                    species_dict[species_key]["valence"] = float(
                                        elements[3].text
                                    )
        d["species_dict"] = species_dict
        species_list = list()
        for key, val in species_dict.items():
            for sp in np.tile([key], species_dict[key]["n_atoms"]):
                species_list.append(clean_character(sp))
        d["species_list"] = species_list

    def parse_fermi_level_to_dict(self, node, d):
        """
        Parses fermi level from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "dos"):
            raise AssertionError()
        for item in node:
            if item.tag == "i":
                self.parse_item_to_dict(item, d)

    def parse_total_dos_to_dict(self, node, d):
        """
        Parses total dos data from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "total"):
            raise AssertionError()
        for item in node:
            if item.tag == "array":
                for ii in item:
                    if ii.tag == "set":
                        spin_dos_energies = list()
                        spin_dos_density = list()
                        spin_dos_idensity = list()
                        for sp in ii:
                            if sp.tag == "set" and "spin" in sp.attrib["comment"]:
                                try:
                                    values = self._parse_2d_matrix(sp, vec_type=float)
                                    dos_energies = values[:, 0]
                                    dos_density = values[:, 1]
                                    dos_idensity = values[:, 2]
                                    spin_dos_energies.append(dos_energies)
                                    spin_dos_density.append(dos_density)
                                    spin_dos_idensity.append(dos_idensity)
                                except ValueError:
                                    pass
                        d["spin_dos_energies"] = np.array(spin_dos_energies)
                        d["spin_dos_density"] = np.array(spin_dos_density)
                        d["spin_dos_idensity"] = np.array(spin_dos_idensity)

    def parse_partial_dos_to_dict(self, node, d):
        """
        Parses partial dos data from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "partial"):
            raise AssertionError()
        orbital_dict = dict()
        orbital_index = 0
        for item in node:
            if item.tag == "array":
                for ii in item:
                    if ii.tag == "field":
                        if "energy" not in ii.text:
                            orbital_dict[ii.text.replace(" ", "")] = orbital_index
                        orbital_index += 1
                    if ii.tag == "set":
                        atom_resolved_dos = list()
                        for ion in ii:
                            spin_resolved_dos = list()
                            if ion.tag == "set" and "ion" in ion.attrib["comment"]:
                                for sp in ion:
                                    if (
                                        sp.tag == "set"
                                        and "spin" in sp.attrib["comment"]
                                    ):
                                        values = self._parse_2d_matrix(
                                            sp, vec_type=float
                                        )
                                        spin_resolved_dos.append(values[:, 1:])
                            atom_resolved_dos.append(spin_resolved_dos)
                        atom_resolved_dos = np.array(atom_resolved_dos)
                        n_atoms, n_spin, n_densities, n_orbitals = np.shape(
                            atom_resolved_dos
                        )
                        new_grand_dos_matrix = np.zeros(
                            (n_spin, n_atoms, n_orbitals, n_densities)
                        )
                        for i_spin in range(n_spin):
                            for i_atom in range(n_atoms):
                                for i_orbitals in range(n_orbitals):
                                    new_grand_dos_matrix[
                                        i_spin, i_atom, i_orbitals, :
                                    ] = atom_resolved_dos[i_atom, i_spin, :, i_orbitals]
                        d["resolved_dos_matrix"] = new_grand_dos_matrix

    def parse_projected_dos_to_dict(self, node, d):
        """
        Parses partial dos data from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "projected"):
            raise AssertionError()
        orbital_dict = dict()
        orbital_index = 0
        for item in node:
            if item.tag == "array":
                for ii in item:
                    if ii.tag == "field":
                        orbital_dict[ii.text] = orbital_index
                        orbital_index += 1
                    if ii.tag == "set":
                        spin_dos_mat = list()
                        for sp in ii:
                            if sp.tag == "set" and "spin" in sp.attrib["comment"]:
                                kpt_dos_mat = list()
                                for kpt in sp:
                                    band_dos_mat = list()
                                    for band in kpt:
                                        dos_matrix = self._parse_2d_matrix(
                                            band, vec_type=float
                                        )
                                        band_dos_mat.append(dos_matrix)
                                    kpt_dos_mat.append(band_dos_mat)
                                spin_dos_mat.append(kpt_dos_mat)
                        grand_dos_matrix = np.array(spin_dos_mat)
                        d["grand_dos_matrix"] = grand_dos_matrix
                        d["orbital_dict"] = orbital_dict

    def parse_scf(self, node):
        """
        Parses the total energy and dipole moments for a VASP calculation

        Args:
            node: (xml.etree.Element instance): The node to parse

        Returns:
            d (dict): Dictionary to containing parsed data
        """
        d = dict()
        if not (node.tag == "scstep"):
            raise AssertionError()
        for item in node:
            if item.tag == "energy":
                for i in item:
                    if i.attrib["name"] == "e_wo_entrp":
                        d["scf_energy"] = get_float_with_exception(i.text)
                    if i.attrib["name"] == "e_fr_energy":
                        d["scf_fr_energy"] = get_float_with_exception(i.text)
                    if i.attrib["name"] == "e_0_energy":
                        d["scf_0_energy"] = get_float_with_exception(i.text)
            if item.tag == "dipole":
                for i in item:
                    if i.attrib["name"] == "dipole":
                        d["scf_dipole_moment"] = self._parse_vector(i, vec_type=float)
        return d

    def parse_calc_to_dict(self, node, d):
        """
        Parses ionic step data from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        scf_energies = list()
        scf_fr_energies = list()
        scf_0_energies = list()
        scf_moments = list()
        for item in node:
            if item.tag in ["scstep"]:
                scf_dict = self.parse_scf(item)
                scf_energies.append(scf_dict["scf_energy"])
                scf_fr_energies.append(scf_dict["scf_fr_energy"])
                scf_0_energies.append(scf_dict["scf_0_energy"])
                if "scf_dipole_moment" in scf_dict.keys():
                    scf_moments.append(scf_dict["scf_dipole_moment"])
            if item.tag in ["structure"]:
                struct_dict = dict()
                self.parse_structure_to_dict(item, struct_dict)
                d["positions"].append(struct_dict["positions"])
                d["cells"].append(struct_dict["cell"])
            if item.tag in ["varray"] and item.attrib["name"] == "forces":
                d["forces"].append(self._parse_2d_matrix(item, vec_type=float))
            if item.tag in ["stress"]:
                d["stress_tensors"].append(self._parse_2d_matrix(item, vec_type=float))
            if item.tag == "energy":
                for i in item:
                    if i.attrib["name"] == "e_wo_entrp":
                        d["total_energies"].append(float(i.text))
                    if i.attrib["name"] == "e_fr_energy":
                        d["total_fr_energies"].append(float(i.text))
                    if i.attrib["name"] == "e_0_energy":
                        d["total_0_energies"].append(float(i.text))
                    if i.attrib["name"] == "kinetic":
                        d["kinetic_energies"].append(float(i.text))
            if item.tag == "eigenvalues":
                self.parse_eigenvalues_to_dict(item, d)

            if item.tag == "dos":
                self.parse_fermi_level_to_dict(item, d)
                d["efermi"] = float(d["efermi"])
                for i in item:
                    if i.tag == "total":
                        try:
                            self.parse_total_dos_to_dict(i, d)
                        except ValueError:
                            pass
                    if i.tag == "partial":
                        try:
                            self.parse_partial_dos_to_dict(i, d)
                        except ValueError:
                            pass

            if item.tag == "projected":
                self.parse_projected_dos_to_dict(item, d)

            if "cce" in item.tag:
                self.parse_cce_to_dict(item, d)

        d["scf_energies"].append(scf_energies)
        d["scf_fr_energies"].append(scf_fr_energies)
        d["scf_0_energies"].append(scf_0_energies)
        d["scf_dipole_moments"].append(scf_moments)

    @staticmethod
    def parse_cce_to_dict(node, d):
        for item in node:
            if item.attrib["name"] not in list(d.keys()):
                d[item.attrib["name"]] = list()
            d[item.attrib["name"]].append(float(item.text))

    def parse_eigenvalues_to_dict(self, node, d):
        """
        Parses eigenvalue and occupancy data from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "eigenvalues"):
            raise AssertionError()
        grand_eigenvalue_matrix = list()
        grand_occupancy_matrix = list()
        for item in node:
            if item.tag == "array":
                for ii in item:
                    if ii.tag == "set":
                        spin_occ_mat = list()
                        spin_eig_mat = list()
                        for sp in ii:
                            if sp.tag == "set" and "spin" in sp.attrib["comment"]:
                                kpt_eig_mat = list()
                                kpt_occ_mat = list()
                                for kpt in sp:
                                    values = self._parse_2d_matrix(kpt, vec_type=float)
                                    eig_vec = values[:, 0].flatten()
                                    occ_vec = values[:, 1].flatten()
                                    kpt_eig_mat.append(eig_vec)
                                    kpt_occ_mat.append(occ_vec)
                                spin_eig_mat.append(kpt_eig_mat)
                                spin_occ_mat.append(kpt_occ_mat)
                        grand_eigenvalue_matrix = np.array(spin_eig_mat)
                        grand_occupancy_matrix = np.array(spin_occ_mat)
        d["grand_eigenvalue_matrix"] = grand_eigenvalue_matrix
        d["grand_occupancy_matrix"] = grand_occupancy_matrix

    def parse_structure_to_dict(self, node, d):
        """
        Parses structure from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "structure"):
            raise AssertionError()
        for leaf in node:
            if leaf.tag == "crystal":
                for item in leaf:
                    if item.tag == "varray" and item.attrib["name"] == "basis":
                        d["cell"] = self._parse_2d_matrix(item)
            if leaf.tag == "varray" and leaf.attrib["name"] == "positions":
                d["positions"] = self._parse_2d_matrix(leaf)

            if leaf.tag == "varray" and leaf.attrib["name"] == "selective":
                d["selective_dynamics"] = self._parse_2d_matrix(leaf, vec_type=bool)

    @staticmethod
    def parse_item_to_dict(node, d):
        """
        Parses values from an item to a dictionary

        Args:
            node (etree.Element instance): Node to be parsed
            d (dict): The dictionary to which data is to be parsed

        Returns:
            d (dictionary)
        """
        type_dict = {"string": str, "float": float, "int": int, "logical": bool}
        logical_dict = {"T": True, "F": False}
        try:
            if node.attrib["type"] == "logical":
                d[node.attrib["name"]] = logical_dict[node.text.strip()]
            else:
                d[node.attrib["name"]] = type_dict[node.attrib["type"]](node.text)
        except (KeyError, IndexError, ValueError):
            d[node.attrib["name"]] = node.text
        return d

    def parse_parameters(self, node, d):
        """
        Parses parameter data from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
        """
        if not (node.tag == "parameters"):
            raise AssertionError()
        self.parse_recursively(node, d, key_name="parameters")

    def parse_recursively(self, node, d, key_name=None):
        """
        Parses recursively from a node to a dictionary

        Args:
            node (xml.etree.Element instance): The node to parse
            d (dict): The dictionary to which data is to be parsed
            key_name (str): Forcefully assign a key name in case it is not present in the xml file
        """
        if not len(node) > 0:
            d[clean_key(node.attrib["name"])] = clean_character(node.text)
            return
        else:
            try:
                if key_name is not None:
                    dict_key = clean_key(key_name)
                else:
                    dict_key = clean_key(node.attrib["name"])
                d[dict_key] = dict()
                for item in node:
                    try:
                        self.parse_item_to_dict(item, d[dict_key][item.attrib["name"]])
                    except (KeyError, ValueError, IndexError):
                        try:
                            self.parse_recursively(
                                item, d[dict_key], item.attrib["name"]
                            )
                        except (KeyError, ValueError, IndexError):
                            pass
            except KeyError:
                pass

    def _parse_2d_matrix(self, node, vec_type=float):
        """
        Parses a 2D vector from a node

        Args:
            node (xml.etree.Element instance): The node to parse
            vec_type (type): The type of the vector to be parsed

        Returns:
            numpy.ndarray: The required 2D array/vector
        """
        arr = list()
        for item in node:
            arr.append(self._parse_vector(item, vec_type=vec_type))
        return np.array(arr)

    @staticmethod
    def _parse_vector(node, vec_type=float):
        """
        Parses a 1D vector from a node

        Args:
            node (xml.etree.Element instance): The node to parse
            vec_type (type): The type of the vector to be parsed

        Returns:
            numpy.ndarray: The required 1D array/vector
        """
        txt = node.text
        lst = txt.split()
        logical_dict = {"T": True, "F": False}
        if "type" in node.attrib.keys():
            if node.attrib["type"] == "logical":
                return np.array([logical_dict[l.strip()] for l in lst])
            else:
                return np.array([vec_type(l) for l in lst])
        else:
            return np.array([vec_type(l) for l in lst])

    def get_initial_structure(self):
        """
        Gets the initial structure from the simulation

        Returns:
            pyiron.atomistics.structure.atoms.Atoms: The initial structure

        """
        # try:
        el_list = [
            re.split("[^a-zA-Z]", el)[0]
            for el in self.vasprun_dict["atominfo"]["species_list"]
        ]
        cell = self.vasprun_dict["init_structure"]["cell"]
        positions = self.vasprun_dict["init_structure"]["positions"]
        if len(positions[positions > 1.01]) > 0:
            basis = Atoms(el_list, positions=positions, cell=cell, pbc=True)
        else:
            basis = Atoms(el_list, scaled_positions=positions, cell=cell, pbc=True)
        if "selective_dynamics" in self.vasprun_dict["init_structure"].keys():
            constraints_dict = {
                label: []
                for label in ["TTT", "TTF", "FTT", "TFT", "TFF", "FFT", "FTF", "FFF"]
            }
            for i, val in enumerate(
                self.vasprun_dict["init_structure"]["selective_dynamics"]
            ):
                if val[0] and val[1] and val[2]:
                    constraints_dict["TTT"].append(i)
                elif val[0] and val[1] and not val[2]:
                    constraints_dict["TTF"].append(i)
                elif not val[0] and val[1] and val[2]:
                    constraints_dict["FTT"].append(i)
                elif val[0] and not val[1] and val[2]:
                    constraints_dict["TFT"].append(i)
                elif val[0] and not val[1] and not val[2]:
                    constraints_dict["TFF"].append(i)
                elif not val[0] and val[1] and not val[2]:
                    constraints_dict["FTF"].append(i)
                elif not val[0] and not val[1] and val[2]:
                    constraints_dict["FFT"].append(i)
                elif not val[0] and not val[1] and not val[2]:
                    constraints_dict["FFF"].append(i)
                else:
                    raise ValueError("Selective Dynamics Error: " + str(val))

            constraints_lst = []
            for k, v in constraints_dict.items():
                if len(v) > 0:
                    if k == "TTT":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(False, False, False))
                        )
                    elif k == "TTF":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(False, False, True))
                        )
                    elif k == "FTT":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(True, False, False))
                        )
                    elif k == "TFT":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(False, True, False))
                        )
                    elif k == "TFF":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(False, True, True))
                        )
                    elif k == "FTF":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(True, False, True))
                        )
                    elif k == "FFT":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(True, True, False))
                        )
                    elif k == "FFF":
                        constraints_lst.append(
                            FixCartesian(a=v, mask=(True, True, True))
                        )
                    else:
                        raise ValueError(
                            "Selective Dynamics Error: " + str(k) + ": " + str(v)
                        )

            basis.set_constraint(constraints_lst)
        return basis

    def get_final_structure(self):
        """
        Gets the final structure from the simulation

        Returns:
            pyiron.atomistics.structure.atoms.Atoms: The final structure

        """
        # try:
        basis = self.get_initial_structure()
        basis.set_cell(self.vasprun_dict["final_structure"]["cell"])
        positions = self.vasprun_dict["final_structure"]["positions"]
        if len(positions[positions > 1.01]) > 0:
            basis.positions = positions
        else:
            basis.set_scaled_positions(positions)
        return basis
        # except (KeyError, AttributeError, ValueError):
        #     return

    def get_electronic_structure(self, es_class=ElectronicStructure):
        """
        Get's the electronic structure from the VASP calculation

        Returns:
            pyiron.atomistics.waves.electronic.ElectronicStructure: The electronic structure object

        """
        es_obj = es_class()
        es_obj.kpoint_list = self.vasprun_dict["kpoints"]["kpoint_list"]
        es_obj.kpoint_weights = self.vasprun_dict["kpoints"]["kpoint_weights"]
        es_obj.eigenvalue_matrix = self.vasprun_dict["grand_eigenvalue_matrix"]
        es_obj.occupancy_matrix = self.vasprun_dict["grand_occupancy_matrix"]
        es_obj.n_spins = len(es_obj.occupancy_matrix)
        if "grand_dos_matrix" in self.vasprun_dict.keys():
            es_obj.grand_dos_matrix = self.vasprun_dict["grand_dos_matrix"]
        if "efermi" in self.vasprun_dict.keys():
            es_obj.efermi = self.vasprun_dict["efermi"]
        if "spin_dos_energies" in self.vasprun_dict.keys():
            es_obj.dos_energies = self.vasprun_dict["spin_dos_energies"][0]
            es_obj.dos_densities = self.vasprun_dict["spin_dos_density"]
            es_obj.dos_idensities = self.vasprun_dict["spin_dos_idensity"]
        if "resolved_dos_matrix" in self.vasprun_dict.keys():
            es_obj.resolved_densities = self.vasprun_dict["resolved_dos_matrix"]
            es_obj.orbital_dict = self.vasprun_dict["orbital_dict"]
        es_obj.generate_from_matrices()
        return es_obj

    def get_potentiostat_output(self):
        """
        Gets the input from the vasp potentiostat output and stores it using intuitive names

        Returns:
            dict/None: Dictionary with potentiostat output

        """
        potstat_dict = dict()
        if "dftnw_pot" not in self.vasprun_dict.keys():
            return
        try:
            vasprun_dict_keys = [
                "dftnw_pot",
                "dftnw_zval",
                "dftnw_electrodecharge",
                "dftnw_vaclevel_upperside_surf",
                "dftnw_vaclevel_lowerside_surf",
                "dftnw_efermi",
            ]
            potstat_dict_keys = [
                "potential_drop",
                "Ne_charge",
                "electrode_charge",
                "vac_level_upper",
                "vac_level_lower",
                "fermi_level",
            ]
            for k, vk in zip(potstat_dict_keys, vasprun_dict_keys):
                if vk in self.vasprun_dict.keys():
                    potstat_dict[k] = np.array(self.vasprun_dict[vk])
            return potstat_dict
        except KeyError:
            if len(potstat_dict.keys()) > 0:
                return potstat_dict
            else:
                return

    def get_valence_electrons_per_atom(self):
        """
        Gets the number of valence electrons for each atom in the final structure

        Returns:
            numpy.ndarray: Array of valence charges
        """
        sp_dict = self.vasprun_dict["atominfo"]["species_dict"]
        return np.hstack(
            [[val["valence"]] * val["n_atoms"] for val in sp_dict.values()]
        )


def clean_character(a, remove_char=" "):
    """
    Args:
        a (str): String to be cleaned
        remove_char (str): Character to be replaced

    Returns:
        str: The clean string
    """
    if isinstance(a, str):
        return a.replace(remove_char, "")
    else:
        return a


def clean_key(a, remove_char=" "):
    """
    Replaces blanck spaces from a string for a dictionary key with "_"

    Args:
        a (str): String to be cleaned
        remove_char (str): Character to be replaced

    Returns:
        str: The clean string
    """
    if isinstance(a, str):
        return a.replace(remove_char, "_")
    else:
        return a


def get_float_with_exception(text, exception_value=0.0):
    """
    Converts a text into a corresponding float or returns `exception_value` if it can't do this

    Args:
        text (str/numpy.str_): String to convert to float
        exception_value (float/None): Value to be returned if you can't text to a float

    Returns:
        float/None: Exception value

    """
    try:
        return float(text)
    except ValueError:
        warnings.warn(message=" ", category=VasprunWarning)
        return exception_value


class VasprunError(ValueError):
    pass


class VasprunWarning(UserWarning):
    pass



from __future__ import print_function
import os
import posixpath
import numpy as np

from ase.atoms import Atoms
from pyiron_vasp.vasp.parser.outcar import Outcar, OutcarCollectError
from pyiron_vasp.vasp.parser.oszicar import Oszicar
from pyiron_vasp.vasp.procar import Procar
from pyiron_vasp.vasp.structure import read_atoms, vasp_sorter
from pyiron_vasp.vasp.vasprun import Vasprun as Vr
from pyiron_vasp.vasp.vasprun import VasprunError, VasprunWarning
from pyiron_vasp.vasp.volumetric_data import VaspVolumetricData
from pyiron_vasp.dft.bader import Bader
from pyiron_vasp.dft.waves.electronic import ElectronicStructure
import warnings


class Output:
    """
    Handles the output from a VASP simulation.

    Attributes:
        electronic_structure: Gives the electronic structure of the system
        electrostatic_potential: Gives the electrostatic/local potential of the system
        charge_density: Gives the charge density of the system
    """

    def __init__(self):
        self._structure = None
        self.outcar = Outcar()
        self.oszicar = Oszicar()
        self.generic_output = GenericOutput()
        self.description = (
            "This contains all the output static from this particular vasp run"
        )
        self.charge_density = VaspVolumetricData()
        self.electrostatic_potential = VaspVolumetricData()
        self.procar = Procar()
        self.electronic_structure = ElectronicStructure()
        self.vp_new = Vr()

    @property
    def structure(self):
        """
        Getter for the output structure
        """
        return self._structure

    @structure.setter
    def structure(self, atoms):
        """
        Setter for the output structure
        """
        self._structure = atoms

    def collect(
        self, directory=os.getcwd(), sorted_indices=None, es_class=ElectronicStructure
    ):
        """
        Collects output from the working directory

        Args:
            directory (str): Path to the directory
            sorted_indices (np.array/None):
        """
        if sorted_indices is None:
            sorted_indices = vasp_sorter(self.structure)
        files_present = os.listdir(directory)
        log_dict = dict()
        vasprun_working, outcar_working = False, False
        if not ("OUTCAR" in files_present or "vasprun.xml" in files_present):
            raise IOError("Either the OUTCAR or vasprun.xml files need to be present")
        if "OSZICAR" in files_present:
            self.oszicar.from_file(filename=posixpath.join(directory, "OSZICAR"))
        if "OUTCAR" in files_present:
            try:
                self.outcar.from_file(filename=posixpath.join(directory, "OUTCAR"))
                outcar_working = True
            except OutcarCollectError as e:
                warnings.warn(f"OUTCAR present, but could not be parsed: {e}!")
                outcar_working = False
        if "vasprun.xml" in files_present:
            try:
                with warnings.catch_warnings(record=True) as w:
                    warnings.simplefilter("always")
                    self.vp_new.from_file(
                        filename=posixpath.join(directory, "vasprun.xml")
                    )
                    if any([isinstance(warn.category, VasprunWarning) for warn in w]):
                        warnings.warn(
                            "vasprun.xml parsed but with some inconsistencies. "
                            "Check vasp output to be sure",
                            VasprunWarning,
                        )
            except VasprunError:
                warnings.warn(
                    "Unable to parse the vasprun.xml file. Will attempt to get data from OUTCAR"
                )
            else:
                # If parsing the vasprun file does not throw an error, then set to True
                vasprun_working = True
        if outcar_working:
            log_dict["temperature"] = self.outcar.parse_dict["temperatures"]
            log_dict["stresses"] = self.outcar.parse_dict["stresses"]
            log_dict["pressures"] = self.outcar.parse_dict["pressures"]
            log_dict["elastic_constants"] = self.outcar.parse_dict["elastic_constants"]
            self.generic_output.dft_log_dict["n_elect"] = self.outcar.parse_dict[
                "n_elect"
            ]
            if len(self.outcar.parse_dict["magnetization"]) > 0:
                magnetization = np.array(
                    self.outcar.parse_dict["magnetization"], dtype=object
                )
                final_magmoms = np.array(
                    self.outcar.parse_dict["final_magmoms"], dtype=object
                )
                # magnetization[sorted_indices] = magnetization.copy()
                if len(final_magmoms) != 0:
                    if len(final_magmoms.shape) == 3:
                        final_magmoms[:, sorted_indices, :] = final_magmoms.copy()
                    else:
                        final_magmoms[:, sorted_indices] = final_magmoms.copy()
                self.generic_output.dft_log_dict["magnetization"] = (
                    magnetization.tolist()
                )
                self.generic_output.dft_log_dict["final_magmoms"] = (
                    final_magmoms.tolist()
                )
            self.generic_output.dft_log_dict["e_fermi_list"] = self.outcar.parse_dict[
                "e_fermi_list"
            ]
            self.generic_output.dft_log_dict["vbm_list"] = self.outcar.parse_dict[
                "vbm_list"
            ]
            self.generic_output.dft_log_dict["cbm_list"] = self.outcar.parse_dict[
                "cbm_list"
            ]

        if vasprun_working:
            log_dict["forces"] = self.vp_new.vasprun_dict["forces"]
            log_dict["cells"] = self.vp_new.vasprun_dict["cells"]
            log_dict["volume"] = np.linalg.det(self.vp_new.vasprun_dict["cells"])
            # The vasprun parser also returns the energies printed again after the final SCF cycle under the key
            # "total_energies", but due to a bug in the VASP output, the energies reported there are wrong in Vasp 5.*;
            # instead use the last energy from the scf cycle energies
            # BUG link: https://ww.vasp.at/forum/viewtopic.php?p=19242
            try:
                # bug report is not specific to which Vasp5 versions are affected; be safe and workaround for all of
                # them
                is_vasp5 = self.vp_new.vasprun_dict["generator"]["version"].startswith(
                    "5."
                )
            except KeyError:  # in case the parser didn't read the version info
                is_vasp5 = True
            if is_vasp5:
                log_dict["energy_pot"] = np.array(
                    [e[-1] for e in self.vp_new.vasprun_dict["scf_fr_energies"]]
                )
            else:
                # total energies refers here to the total energy of the electronic system, not the total system of
                # electrons plus (potentially) moving ions; hence this is the energy_pot
                log_dict["energy_pot"] = self.vp_new.vasprun_dict["total_fr_energies"]
            if "kinetic_energies" in self.vp_new.vasprun_dict.keys():
                log_dict["energy_tot"] = (
                    log_dict["energy_pot"]
                    + self.vp_new.vasprun_dict["kinetic_energies"]
                )
            else:
                log_dict["energy_tot"] = log_dict["energy_pot"]
            log_dict["steps"] = np.arange(len(log_dict["energy_tot"]))
            log_dict["positions"] = self.vp_new.vasprun_dict["positions"]
            log_dict["forces"][:, sorted_indices] = log_dict["forces"].copy()
            log_dict["positions"][:, sorted_indices] = log_dict["positions"].copy()
            log_dict["positions"] = np.einsum(
                "nij,njk->nik", log_dict["positions"], log_dict["cells"]
            )
            # log_dict["scf_energies"] = self.vp_new.vasprun_dict["scf_energies"]
            # log_dict["scf_dipole_moments"] = self.vp_new.vasprun_dict["scf_dipole_moments"]
            self.electronic_structure = self.vp_new.get_electronic_structure(
                es_class=es_class,
            )
            if self.electronic_structure.grand_dos_matrix is not None:
                self.electronic_structure.grand_dos_matrix[
                    :, :, :, sorted_indices, :
                ] = self.electronic_structure.grand_dos_matrix[:, :, :, :, :].copy()
            if self.electronic_structure.resolved_densities is not None:
                self.electronic_structure.resolved_densities[
                    :, sorted_indices, :, :
                ] = self.electronic_structure.resolved_densities[:, :, :, :].copy()
            self.structure.positions = log_dict["positions"][-1]
            self.structure.set_cell(log_dict["cells"][-1])
            self.generic_output.dft_log_dict["potentiostat_output"] = (
                self.vp_new.get_potentiostat_output()
            )
            valence_charges_orig = self.vp_new.get_valence_electrons_per_atom()
            valence_charges = valence_charges_orig.copy()
            valence_charges[sorted_indices] = valence_charges_orig
            self.generic_output.dft_log_dict["valence_charges"] = valence_charges

        elif outcar_working:
            # log_dict = self.outcar.parse_dict.copy()
            if len(self.outcar.parse_dict["energies"]) == 0:
                raise VaspCollectError("Error in parsing OUTCAR")
            log_dict["energy_tot"] = self.outcar.parse_dict["energies"]
            log_dict["temperature"] = self.outcar.parse_dict["temperatures"]
            log_dict["stresses"] = self.outcar.parse_dict["stresses"]
            log_dict["pressures"] = self.outcar.parse_dict["pressures"]
            log_dict["forces"] = self.outcar.parse_dict["forces"]
            log_dict["positions"] = self.outcar.parse_dict["positions"]
            log_dict["forces"][:, sorted_indices] = log_dict["forces"].copy()
            log_dict["positions"][:, sorted_indices] = log_dict["positions"].copy()
            if len(log_dict["positions"].shape) != 3:
                raise VaspCollectError("Improper OUTCAR parsing")
            elif log_dict["positions"].shape[1] != len(sorted_indices):
                raise VaspCollectError("Improper OUTCAR parsing")
            if len(log_dict["forces"].shape) != 3:
                raise VaspCollectError("Improper OUTCAR parsing")
            elif log_dict["forces"].shape[1] != len(sorted_indices):
                raise VaspCollectError("Improper OUTCAR parsing")
            log_dict["time"] = self.outcar.parse_dict["time"]
            log_dict["steps"] = self.outcar.parse_dict["steps"]
            log_dict["cells"] = self.outcar.parse_dict["cells"]
            log_dict["volume"] = np.array(
                [np.linalg.det(cell) for cell in self.outcar.parse_dict["cells"]]
            )
            self.generic_output.dft_log_dict["scf_energy_free"] = (
                self.outcar.parse_dict["scf_energies"]
            )
            self.generic_output.dft_log_dict["scf_dipole_mom"] = self.outcar.parse_dict[
                "scf_dipole_moments"
            ]
            self.generic_output.dft_log_dict["n_elect"] = self.outcar.parse_dict[
                "n_elect"
            ]
            self.generic_output.dft_log_dict["energy_int"] = self.outcar.parse_dict[
                "energies_int"
            ]
            self.generic_output.dft_log_dict["energy_free"] = self.outcar.parse_dict[
                "energies"
            ]
            self.generic_output.dft_log_dict["energy_zero"] = self.outcar.parse_dict[
                "energies_zero"
            ]
            self.generic_output.dft_log_dict["energy_int"] = self.outcar.parse_dict[
                "energies_int"
            ]
            if "PROCAR" in files_present:
                try:
                    self.electronic_structure = self.procar.from_file(
                        filename=posixpath.join(directory, "PROCAR")
                    )
                    #  Even the atom resolved values have to be sorted from the vasp atoms order to the Atoms order
                    self.electronic_structure.grand_dos_matrix[
                        :, :, :, sorted_indices, :
                    ] = self.electronic_structure.grand_dos_matrix[:, :, :, :, :].copy()
                    try:
                        self.electronic_structure.efermi = self.outcar.parse_dict[
                            "fermi_level"
                        ]
                    except KeyError:
                        self.electronic_structure.efermi = self.vp_new.vasprun_dict[
                            "efermi"
                        ]
                except ValueError:
                    pass
        # important that we "reverse sort" the atoms in the vasp format into the atoms in the atoms class
        self.generic_output.log_dict = log_dict
        if vasprun_working:
            # self.dft_output.log_dict["parameters"] = self.vp_new.vasprun_dict["parameters"]
            self.generic_output.dft_log_dict["scf_dipole_mom"] = (
                self.vp_new.vasprun_dict["scf_dipole_moments"]
            )
            if len(self.generic_output.dft_log_dict["scf_dipole_mom"][0]) > 0:
                total_dipole_moments = np.array(
                    [
                        dip[-1]
                        for dip in self.generic_output.dft_log_dict["scf_dipole_mom"]
                    ]
                )
                self.generic_output.dft_log_dict["dipole_mom"] = total_dipole_moments
            self.generic_output.dft_log_dict["scf_energy_int"] = (
                self.vp_new.vasprun_dict["scf_energies"]
            )
            self.generic_output.dft_log_dict["scf_energy_free"] = (
                self.vp_new.vasprun_dict["scf_fr_energies"]
            )
            self.generic_output.dft_log_dict["scf_energy_zero"] = (
                self.vp_new.vasprun_dict["scf_0_energies"]
            )
            self.generic_output.dft_log_dict["energy_int"] = np.array(
                [
                    e_int[-1]
                    for e_int in self.generic_output.dft_log_dict["scf_energy_int"]
                ]
            )
            self.generic_output.dft_log_dict["energy_free"] = np.array(
                [
                    e_free[-1]
                    for e_free in self.generic_output.dft_log_dict["scf_energy_free"]
                ]
            )
            # Overwrite energy_free with much better precision from the OSZICAR file
            if "energy_pot" in self.oszicar.parse_dict.keys():
                if np.array_equal(
                    self.generic_output.dft_log_dict["energy_free"],
                    np.round(self.oszicar.parse_dict["energy_pot"], 8),
                ):
                    self.generic_output.dft_log_dict["energy_free"] = (
                        self.oszicar.parse_dict["energy_pot"]
                    )
            self.generic_output.dft_log_dict["energy_zero"] = np.array(
                [
                    e_zero[-1]
                    for e_zero in self.generic_output.dft_log_dict["scf_energy_zero"]
                ]
            )
            self.generic_output.dft_log_dict["n_elect"] = float(
                self.vp_new.vasprun_dict["parameters"]["electronic"]["NELECT"]
            )
            if "kinetic_energies" in self.vp_new.vasprun_dict.keys():
                # scf_energy_kin is for backwards compatibility
                self.generic_output.dft_log_dict["scf_energy_kin"] = (
                    self.vp_new.vasprun_dict["kinetic_energies"]
                )
                self.generic_output.dft_log_dict["energy_kin"] = (
                    self.vp_new.vasprun_dict["kinetic_energies"]
                )

        if (
            "LOCPOT" in files_present
            and os.stat(posixpath.join(directory, "LOCPOT")).st_size != 0
        ):
            self.electrostatic_potential.from_file(
                filename=posixpath.join(directory, "LOCPOT"), normalize=False
            )
        if (
            "CHGCAR" in files_present
            and os.stat(posixpath.join(directory, "CHGCAR")).st_size != 0
        ):
            self.charge_density.from_file(
                filename=posixpath.join(directory, "CHGCAR"), normalize=True
            )
        self.generic_output.bands = self.electronic_structure

    def to_dict(self):
        output_dict = {
            "description": self.description,
            "generic": self.generic_output.to_dict(),
        }

        if self.structure is not None and type(self.structure) == Atoms:
            output_dict["structure"] = self.structure.todict()
        elif self.structure is not None:
            output_dict["structure"] = self.structure.to_dict()

        if self.electrostatic_potential.total_data is not None:
            output_dict["electrostatic_potential"] = (
                self.electrostatic_potential.to_dict()
            )

        if self.charge_density.total_data is not None:
            output_dict["charge_density"] = self.charge_density.to_dict()

        if len(self.electronic_structure.kpoint_list) > 0:
            output_dict["electronic_structure"] = self.electronic_structure.to_dict()

        if len(self.outcar.parse_dict.keys()) > 0:
            output_dict["outcar"] = self.outcar.to_dict_minimal()
        return output_dict


class GenericOutput:
    """

    This class stores the generic output like different structures, energies and forces from a simulation in a highly
    generic format. Usually the user does not have to access this class.

    Attributes:
        log_dict (dict): A dictionary of all tags and values of generic data (positions, forces, etc)
    """

    def __init__(self):
        self.log_dict = dict()
        self.dft_log_dict = dict()
        self.description = "generic_output contains generic output static"
        self._bands = ElectronicStructure()

    @property
    def bands(self):
        return self._bands

    @bands.setter
    def bands(self, val):
        self._bands = val

    def to_dict(self):
        go_dict, dft_dict = {}, {}
        for key, val in self.log_dict.items():
            go_dict[key] = val
        for key, val in self.dft_log_dict.items():
            dft_dict[key] = val
        go_dict["dft"] = dft_dict
        if self.bands.eigenvalue_matrix is not None:
            go_dict["dft"]["bands"] = self.bands.to_dict()
        return go_dict


class VaspCollectError(ValueError):
    pass


def parse_vasp_output(
    working_directory: str,
    structure: Atoms = None,
    sorted_indices: list = None,
    read_atoms_funct: callable = read_atoms,
    es_class=ElectronicStructure,
    bader_class=Bader,
    output_parser_class=Output,
) -> dict:
    """
    Parse the VASP output in the working_directory and return it as hierachical dictionary.

    Args:
        working_directory (str): directory of the VASP calculation
        structure (Atoms): atomistic structure as optional input for matching the output to the input of the calculation
        sorted_indices (list): list of indices used to sort the atomistic structure

    Returns:
        dict: hierarchical output dictionary
    """
    output_parser = output_parser_class()
    if structure is None or len(structure) == 0:
        try:
            structure = get_final_structure_from_file(
                working_directory=working_directory,
                filename="CONTCAR",
                read_atoms_funct=read_atoms_funct,
            )
        except IOError:
            structure = get_final_structure_from_file(
                working_directory=working_directory,
                filename="POSCAR",
                read_atoms_funct=read_atoms_funct,
            )
    if sorted_indices is None:
        sorted_indices = np.array(range(len(structure)))
    output_parser.structure = structure.copy()
    try:
        output_parser.collect(
            directory=working_directory,
            sorted_indices=sorted_indices,
            es_class=es_class,
        )
    except VaspCollectError:
        raise
    # Try getting high precision positions from CONTCAR
    try:
        output_parser.structure = get_final_structure_from_file(
            working_directory=working_directory,
            filename="CONTCAR",
            structure=structure,
            sorted_indices=sorted_indices,
            read_atoms_funct=read_atoms_funct,
        )
    except (IOError, ValueError, FileNotFoundError):
        pass

    # Bader analysis
    if os.path.isfile(os.path.join(working_directory, "AECCAR0")) and os.path.isfile(
        os.path.join(working_directory, "AECCAR2")
    ):
        bader = bader_class(working_directory=working_directory, structure=structure)
        try:
            charges_orig, volumes_orig = bader.compute_bader_charges()
        except ValueError:
            warnings.warn("Invoking Bader charge analysis failed")
        else:
            charges, volumes = charges_orig.copy(), volumes_orig.copy()
            charges[sorted_indices] = charges_orig
            volumes[sorted_indices] = volumes_orig
            if "valence_charges" in output_parser.generic_output.dft_log_dict.keys():
                valence_charges = output_parser.generic_output.dft_log_dict[
                    "valence_charges"
                ]
                # Positive values indicate electron depletion
                output_parser.generic_output.dft_log_dict["bader_charges"] = (
                    valence_charges - charges
                )
            output_parser.generic_output.dft_log_dict["bader_volumes"] = volumes
    return output_parser.to_dict()


def get_final_structure_from_file(
    working_directory,
    filename="CONTCAR",
    structure=None,
    sorted_indices=None,
    read_atoms_funct=read_atoms,
):
    """
    Get the final structure of the simulation usually from the CONTCAR file

    Args:
        filename (str): Path to the CONTCAR file in VASP

    Returns:
        pyiron.atomistics.structure.atoms.Atoms: The final structure
    """
    filename = posixpath.join(working_directory, filename)
    if structure is not None and sorted_indices is None:
        sorted_indices = vasp_sorter(structure)
    if structure is None:
        try:
            output_structure = read_atoms_funct(filename=filename)
            input_structure = output_structure.copy()
        except (IndexError, ValueError, IOError):
            raise IOError("Unable to read output structure")
    else:
        input_structure = structure.copy()
        if type(input_structure) == Atoms:
            species_list = input_structure.get_chemical_symbols()
        else:
            species_list = input_structure.get_parent_symbols()
        try:
            output_structure = read_atoms_funct(
                filename=filename,
                species_list=species_list,
            )
            input_structure.cell = output_structure.cell.copy()
            input_structure.positions[sorted_indices] = output_structure.positions
        except (IndexError, ValueError, IOError):
            raise IOError("Unable to read output structure")
    return input_structure



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from ase.atoms import Atoms
from pyiron_vasp.vasp.structure import write_poscar

__author__ = "Sudarsan Surendralal, Su-Hyun Yoo"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH "
    "- Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class VolumetricData(object):
    """
    A new class to handle 3-dimensional volumetric data elegantly (charge densities, electrostatic potentials etc) based
    on the numpy.ndarray instance. This module is adapted from the pymatgen vasp VolumtricData class

    http://pymatgen.org/_modules/pymatgen/io/vasp/outputs.html#VolumetricData

    Attributes:

        total_data (numpy.ndarray): A 3D array containing the data

    """

    def __init__(self):
        self._total_data = None
        self._atoms = None

    @property
    def atoms(self):
        """
        The structure related to the volumeric data

        Returns:
            pyiron_atomistics.atomistics.structure.Atoms: The structure associated with the data

        """
        return self._atoms

    @atoms.setter
    def atoms(self, val):
        self._atoms = val

    @property
    def total_data(self):
        """
        numpy.ndarray: The Nx x Ny x Nz sized array for the total data
        """
        return self._total_data

    @total_data.setter
    def total_data(self, val):
        if not (isinstance(val, (np.ndarray, list))):
            raise TypeError(
                "Attribute total_data should be a numpy.ndarray instance or a list and "
                "not {}".format(type(val))
            )
        val = np.array(val)
        shape = np.array(np.shape(val))
        if not (len(shape) == 3):
            raise ValueError("Attribute total_data should be a 3D array")
        self._total_data = val

    @staticmethod
    def gauss_f(d, fwhm=0.529177):
        """
        Generates a Gaussian distribution for a given distance and full width half maximum value

        Args:
            d (float): distance between target point and reference point
            fwhm (float): Full width half maximum in angstrom

        Returns:
            float: Gaussian reduction constant

        """
        sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
        d2 = d * d
        return np.exp(-1 / (2 * sigma**2) * d2)

    @staticmethod
    def dist_between_two_grid_points(
        target_grid_point, n_grid_at_center, lattice, grid_shape
    ):
        """
        Calculates the distance between a target grid point and another grid point

        Args:
            target_grid_point (numpy.ndarray/list): Target grid point
            n_grid_at_center (numpy.ndarray/list): coordinate of center of sphere
            lattice (numpy.ndarray/list): lattice vector
            grid_shape (tuple/list/numpy.ndarray): size of grid

        Returns:

            float: Distance between target grid and center of sphere in angstrom

        """
        unit_dist_in_grid = [
            np.sqrt(np.dot(lattice[0], lattice[0])) / grid_shape[0],
            np.sqrt(np.dot(lattice[1], lattice[1])) / grid_shape[1],
            np.sqrt(np.dot(lattice[2], lattice[2])) / grid_shape[2],
        ]
        dn = np.multiply(
            np.subtract(target_grid_point, n_grid_at_center), unit_dist_in_grid
        )
        dist = np.linalg.norm(dn)
        return dist

    def spherical_average_potential(
        self, structure, spherical_center, rad=2, fwhm=0.529177
    ):
        """
        Calculates the spherical average about a given point in space

        Args:
            structure (pyiron_atomistics.atomistics.structure.Atoms): Input structure
            spherical_center (list/numpy.ndarray): position of spherical_center in direct coordinate
            rad (float): radius of sphere to be considered in Angstrom (recommended value: 2)
            fwhm (float): Full width half maximum of gaussian function in Angstrom (recommended value: 0.529177)

        Returns:
            float: Spherical average at the target center

        """
        grid_shape = self._total_data.shape

        # Position of center of sphere at grid coordinates
        n_grid_at_center = [
            int(np.ceil(spherical_center[0] * grid_shape[0])),
            int(np.ceil(spherical_center[1] * grid_shape[1])),
            int(np.ceil(spherical_center[2] * grid_shape[2])),
        ]

        # Unit distance between grids
        dist_in_grid = [
            np.linalg.norm(structure.cell[0]) / grid_shape[0],
            np.linalg.norm(structure.cell[1]) / grid_shape[1],
            np.linalg.norm(structure.cell[2]) / grid_shape[2],
        ]

        # Range of grids to be considered within the provided radius w.r.t. center of sphere
        num_grid_in_sph = [[], []]
        for i, dist in enumerate(dist_in_grid):
            num_grid_in_sph[0].append(n_grid_at_center[i] - int(np.ceil(rad / dist)))
            num_grid_in_sph[1].append(n_grid_at_center[i] + int(np.ceil(rad / dist)))

        sph_avg_tmp = []
        weight = 0
        for k in range(num_grid_in_sph[0][0], num_grid_in_sph[1][0]):
            for l in range(num_grid_in_sph[0][1], num_grid_in_sph[1][1]):
                for m in range(num_grid_in_sph[0][2], num_grid_in_sph[1][2]):
                    target_grid_point = [k, l, m]
                    dist = self.dist_between_two_grid_points(
                        target_grid_point, n_grid_at_center, structure.cell, grid_shape
                    )
                    if dist <= rad:
                        sph_avg_tmp.append(
                            self._total_data[
                                k % grid_shape[0], l % grid_shape[1], m % grid_shape[2]
                            ]
                            * self.gauss_f(dist, fwhm)
                        )
                        weight += self.gauss_f(dist, fwhm)
                    else:
                        pass
        sum_list = np.sum(sph_avg_tmp)
        sph_avg = sum_list / weight
        return sph_avg

    @staticmethod
    def dist_between_two_grid_points_cyl(
        target_grid_point, n_grid_at_center, lattice, grid_shape, direction_of_cyl
    ):
        """
        Distance between a target grid point and the center of a cylinder

        Args:
            target_grid_point (numpy.ndarray/list): Target grid point
            n_grid_at_center (numpy.ndarray/list): coordinate of center of sphere
            lattice (numpy.ndarray/list): lattice vector
            grid_shape (tuple/list/numpy.ndarray): size of grid
            direction_of_cyl (int): Axis of cylinder (0 (x) or 1 (y) or 2 (z))

        Returns:
            float: Distance between target grid and in-plane center of cylinder

        """
        unit_dist_in_grid = [
            np.sqrt(np.dot(lattice[0], lattice[0])) / grid_shape[0],
            np.sqrt(np.dot(lattice[1], lattice[1])) / grid_shape[1],
            np.sqrt(np.dot(lattice[2], lattice[2])) / grid_shape[2],
        ]
        dn = np.multiply(
            np.subtract(target_grid_point, n_grid_at_center), unit_dist_in_grid
        )
        if direction_of_cyl == 0:
            dn[0] = 0
        elif direction_of_cyl == 1:
            dn[1] = 0
        elif direction_of_cyl == 2:
            dn[2] = 0
        else:
            print("check the direction of cylindrical axis")
        dist = np.linalg.norm(dn)
        return dist

    def cylindrical_average_potential(
        self, structure, spherical_center, axis_of_cyl, rad=2, fwhm=0.529177
    ):
        """
        Calculates the cylindrical average about a given point in space

        Args:
            structure (pyiron_atomistics.atomistics.structure.Atoms): Input structure
            spherical_center (list/numpy.ndarray): position of spherical_center in direct coordinate
            rad (float): radius of sphere to be considered in Angstrom (recommended value: 2)
            fwhm (float): Full width half maximum of gaussian function in Angstrom (recommended value: 0.529177)
            axis_of_cyl (int): Axis of cylinder (0 (x) or 1 (y) or 2 (z))

        Returns:
            float: Cylindrical average at the target center

        """
        grid_shape = self._total_data.shape

        # Position of center of sphere at grid coordinates
        n_grid_at_center = [
            int(np.ceil(spherical_center[0] * grid_shape[0])),
            int(np.ceil(spherical_center[1] * grid_shape[1])),
            int(np.ceil(spherical_center[2] * grid_shape[2])),
        ]

        # Unit distance between grids
        dist_in_grid = [
            np.linalg.norm(structure.cell[0]) / grid_shape[0],
            np.linalg.norm(structure.cell[1]) / grid_shape[1],
            np.linalg.norm(structure.cell[2]) / grid_shape[2],
        ]

        # Range of grids to be considered within the provided radius w.r.t. center of sphere
        num_grid_in_cyl = [[], []]

        for i, dist in enumerate(dist_in_grid):
            if i == axis_of_cyl:
                num_grid_in_cyl[0].append(0)
                num_grid_in_cyl[1].append(grid_shape[i])
            else:
                num_grid_in_cyl[0].append(
                    n_grid_at_center[i] - int(np.ceil(rad / dist))
                )
                num_grid_in_cyl[1].append(
                    n_grid_at_center[i] + int(np.ceil(rad / dist))
                )

        cyl_avg_tmp = []
        weight = 0
        for k in range(num_grid_in_cyl[0][0], num_grid_in_cyl[1][0]):
            for l in range(num_grid_in_cyl[0][1], num_grid_in_cyl[1][1]):
                for m in range(num_grid_in_cyl[0][2], num_grid_in_cyl[1][2]):
                    target_grid_point = [k, l, m]
                    dist = self.dist_between_two_grid_points_cyl(
                        target_grid_point,
                        n_grid_at_center,
                        structure.cell,
                        grid_shape,
                        axis_of_cyl,
                    )
                    if dist <= rad:
                        cyl_avg_tmp.append(
                            self._total_data[
                                k % grid_shape[0], l % grid_shape[1], m % grid_shape[2]
                            ]
                            * self.gauss_f(dist, fwhm)
                        )
                        weight += self.gauss_f(dist, fwhm)
                    else:
                        pass
        sum_list = np.sum(cyl_avg_tmp)
        cyl_avg = sum_list / weight

        return cyl_avg

    def get_average_along_axis(self, ind=2):
        """
        Get the lateral average along a certain axis direction. This function is adapted from the pymatgen vasp
        VolumetricData class

        http://pymatgen.org/_modules/pymatgen/io/vasp/outputs.html#VolumetricData.get_average_along_axis

        Args:
            ind (int): Index of axis (0, 1 and 2 for the x, y, and z axis respectively)

        Returns:
            numpy.ndarray: A 1D vector with the laterally averaged values of the volumetric data
        """
        if ind == 0:
            return np.average(np.average(self._total_data, axis=1), 1)
        elif ind == 1:
            return np.average(np.average(self._total_data, axis=0), 1)
        else:
            return np.average(np.average(self._total_data, axis=0), 0)

    def write_cube_file(self, filename="cube_file.cube", cell_scaling=1.0):
        """
        Write the volumetric data into the CUBE file format

        Args:
            filename (str): Filename
            cell_scaling (float): Scale the cell by this fraction

        """
        if self._atoms is None:
            raise ValueError(
                "The volumetric data object must have a valid structure assigned to it before writing "
                "to the cube format"
            )
        data = self.total_data
        n_x, n_y, _ = data.shape
        origin = np.zeros(3)
        flattened_data = np.hstack(
            [data[i, j, :] for i in range(n_x) for j in range(n_y)]
        )
        n_atoms = len(self.atoms)
        total_lines = int(len(flattened_data) / 6) * 6
        reshaped_data = np.reshape(flattened_data[0:total_lines], (-1, 6))
        last_line = [flattened_data[total_lines:]]
        head_array = np.zeros((4, 4))
        head_array[0] = np.append([n_atoms], origin)
        head_array[1:, 0] = data.shape
        head_array[1:, 1:] = self.atoms.cell / data.shape * cell_scaling
        position_array = np.zeros((len(self.atoms.positions), 5))
        position_array[:, 0] = self.atoms.get_atomic_numbers()
        position_array[:, 2:] = self.atoms.positions
        with open(filename, "w") as f:
            f.write("Cube file generated by pyiron (http://pyiron.org) \n")
            f.write("z is the fastest index \n")
            np.savetxt(f, head_array, fmt="%4d %.6f %.6f %.6f")
            np.savetxt(f, position_array, fmt="%4d %.6f %.6f %.6f %.6f")
            np.savetxt(f, reshaped_data, fmt="%.5e")
            np.savetxt(f, last_line, fmt="%.5e")

    def read_cube_file(self, filename="cube_file.cube"):
        """
        Generate data from a CUBE file

        Args:
            filename (str): Filename to parse

        """
        with open(filename, "r", errors="ignore") as f:
            lines = f.readlines()
            n_atoms = int(lines[2].strip().split()[0])
            cell_data = np.genfromtxt(lines[3:6])
            cell_grid = cell_data[:, 1:]
            grid_shape = np.array(cell_data[:, 0], dtype=int)
            # total_data = np.zeros(grid_shape)
            cell = np.array([val * grid_shape[i] for i, val in enumerate(cell_grid)])
            if n_atoms > 0:
                pos_data = np.genfromtxt(lines[6 : n_atoms + 6])
                if n_atoms == 1:
                    pos_data = np.array([pos_data])
                atomic_numbers = np.array(pos_data[:, 0], dtype=int)
                positions = pos_data[:, 2:]
                self._atoms = Atoms(
                    numbers=atomic_numbers, positions=positions, cell=cell
                )
            end_int = n_atoms + 6 + int(np.prod(grid_shape) / 6)
            data = np.genfromtxt(lines[n_atoms + 6 : end_int])
            data_flatten = np.hstack(data)
            if np.prod(grid_shape) % 6 > 0:
                data_flatten = np.append(
                    data_flatten, [float(val) for val in lines[end_int].split()]
                )
            n_x, n_y, n_z = grid_shape
            self._total_data = data_flatten.reshape((n_x, n_y, n_z))

    def write_vasp_volumetric(self, filename="CHGCAR", normalize=False):
        """
        Writes volumetric data into a VASP CHGCAR format

        Args:
            filename (str): Filename of the new file
            normalize (bool): True if the data is to be normalized by the volume

        """
        write_poscar(structure=self.atoms, filename=filename)
        with open(filename, "a") as f:
            f.write("\n")
            f.write(" ".join(list(np.array(self.total_data.shape, dtype=str))))
            f.write("\n")
            _, n_y, n_z = self.total_data.shape
            flattened_data = np.hstack(
                [self.total_data[:, i, j] for j in range(n_z) for i in range(n_y)]
            )
            if normalize:
                flattened_data /= self.atoms.get_volume()
            num_lines = int(len(flattened_data) / 5) * 5
            reshaped_data = np.reshape(flattened_data[0:num_lines], (-1, 5))
            np.savetxt(f, reshaped_data, fmt="%.12f")
            if len(flattened_data) % 5 > 0:
                np.savetxt(f, [flattened_data[num_lines:]], fmt="%.12f")
